{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 1995, 1996 AO ROSNO             }
{                                                       }
{*******************************************************}

unit DBSecur;

interface

{$IFDEF WIN32}
uses SysUtils, Windows, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DBTables, LoginDlg, ChPswDlg;
{$ELSE}
uses SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, DB, DBTables, LoginDlg, ChPswDlg;
{$ENDIF}

type

{ TDBSecurity }

  TUpdateCaption = (ucNoChange, ucAppTitle, ucFormCaption);

  TDBSecurity = class(TComponent)
  private
    FActive: Boolean;
    FAttemptNumber: Integer;
    FDatabase: TDatabase;
    FUsersTableName: TFileName;
    FLoginNameField: PString;
    FLoggedUser: PString;
    FOnCheckUser: TCheckUserEvent;
    FMaxPasswordLen: Integer;
    FAllowEmpty: Boolean;
    FUpdateCaption: TUpdateCaption;
    FOnChangePassword: TChangePasswordEvent;
    FIniFileName: PString;
    FUseRegistry: Boolean;
    FAfterLogin: TNotifyEvent;
    FSaveOnRestore: TNotifyEvent;
    FOnUnlock: TCheckUnlockEvent;
    procedure SetDatabase(Value: TDatabase);
    procedure SetUsersTableName(const Value: TFileName);
    function GetLoginNameField: string;
    procedure SetLoginNameField(const Value: string);
    function GetLoggedUser: string;
    procedure SetLoggedUser(const Value: string);
    function GetIniFileName: string;
    procedure SetIniFileName(const Value: string);
    function UnlockHook(var Message: TMessage): Boolean;
  protected
    procedure DoAfterLogin; virtual;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function Login: Boolean;
    function ChangePassword: Boolean;
    procedure Lock;
    property LoggedUser: string read GetLoggedUser;
  published
    property Active: Boolean read FActive write FActive default True;
    property AllowEmptyPassword: Boolean read FAllowEmpty write FAllowEmpty default True;
    property AttemptNumber: Integer read FAttemptNumber write FAttemptNumber default 3;
    property Database: TDatabase read FDatabase write SetDatabase;
    property IniFileName: string read GetIniFileName write SetIniFileName;
    property LoginNameField: string read GetLoginNameField write SetLoginNameField;
    property MaxPasswordLen: Integer read FMaxPasswordLen write FMaxPasswordLen default 0;
    property UpdateCaption: TUpdateCaption read FUpdateCaption write FUpdateCaption default ucNoChange;
{$IFDEF WIN32}
    property UseRegistry: Boolean read FUseRegistry write FUseRegistry default False;
{$ENDIF WIN32}
    property UsersTableName: TFileName read FUsersTableName write SetUsersTableName;
    property AfterLogin: TNotifyEvent read FAfterLogin write FAfterLogin;
    property OnCheckUser: TCheckUserEvent read FOnCheckUser write FOnCheckUser;
    property OnChangePassword: TChangePasswordEvent read FOnChangePassword
      write FOnChangePassword;
    property OnUnlock: TCheckUnlockEvent read FOnUnlock write FOnUnlock;
  end;

implementation

uses AppUtils;

{ TDBSecurity }

constructor TDBSecurity.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIniFileName := NullStr;
  FActive := True;
  FAttemptNumber := 3;
  FAllowEmpty := True;
  FUseRegistry := False;
  FLoginNameField := NullStr;
  FLoggedUser := NullStr;
end;

destructor TDBSecurity.Destroy;
begin
  DisposeStr(FLoggedUser);
  DisposeStr(FLoginNameField);
  DisposeStr(FIniFileName);
  inherited Destroy;
end;

procedure TDBSecurity.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = Database) then Database := nil;
end;

procedure TDBSecurity.Loaded;
begin
  inherited Loaded;
  if not (csDesigning in ComponentState) and Active and
    (Database <> nil) then
  begin
    Database.LoginPrompt := True;
    if not Login then Halt(10);
  end;
end;

procedure TDBSecurity.SetDatabase(Value: TDatabase);
begin
  if FDatabase <> Value then begin
    FDatabase := Value;
{$IFDEF WIN32}
    if Value <> nil then Value.FreeNotification(Self);
{$ENDIF}
  end;
end;

procedure TDBSecurity.SetUsersTableName(const Value: TFileName);
begin
  if FUsersTableName <> Value then
    FUsersTableName := Value;
end;

function TDBSecurity.GetIniFileName: string;
begin
  Result := FIniFileName^;
  if (Result = '') and not (csDesigning in ComponentState) then begin
{$IFDEF WIN32}
    if UseRegistry then Result := GetDefaultIniRegKey
    else Result := GetDefaultIniName;
{$ELSE}
    Result := GetDefaultIniName;
{$ENDIF}
  end;
end;

procedure TDBSecurity.SetIniFileName(const Value: string);
begin
  AssignStr(FIniFileName, Value);
end;

function TDBSecurity.GetLoginNameField: string;
begin
  Result := FLoginNameField^;
end;

procedure TDBSecurity.SetLoginNameField(const Value: string);
begin
  AssignStr(FLoginNameField, Value);
end;

function TDBSecurity.GetLoggedUser: string;
begin
  Result := FLoggedUser^;
end;

procedure TDBSecurity.SetLoggedUser(const Value: string);
begin
  AssignStr(FLoggedUser, Value);
end;

procedure TDBSecurity.DoAfterLogin;
begin
  if Assigned(FAfterLogin) then FAfterLogin(Self);
end;

function TDBSecurity.Login: Boolean;
var
  LoginName: string;
  F: TForm;
begin
  LoginName := EmptyStr;
  Result := LoginDialog(Database, AttemptNumber, UsersTableName,
    LoginNameField, MaxPasswordLen, OnCheckUser, LoginName, IniFileName,
    FUseRegistry);
  if Result then begin
    SetLoggedUser(LoginName);
    F := Application.MainForm;
    if (F = nil) and (Owner is TForm) then F := Owner as TForm;
    if (F <> nil) and (LoginName <> '') then
      case UpdateCaption of
        ucAppTitle:
          F.Caption := Format('%s (%s)', [Application.Title, LoginName]);
        ucFormCaption:
          F.Caption := Format('%s (%s)', [F.Caption, LoginName]);
      end;
    DoAfterLogin;
  end;
end;

function TDBSecurity.ChangePassword: Boolean;
begin
  Result := ChangePasswordDialog(Database, AttemptNumber, UsersTableName,
    LoginNameField, LoggedUser, MaxPasswordLen, FAllowEmpty,
    OnChangePassword);
end;

procedure TDBSecurity.Lock;
begin
  FSaveOnRestore := Application.OnRestore;
  Application.HookMainWindow(UnlockHook);
  Application.Minimize;
end;

function TDBSecurity.UnlockHook(var Message: TMessage): Boolean;
begin
  Result := False;
  with Message do begin
    case Msg of
      WM_SYSCOMMAND:
        if WParam and $FFF0 = SC_RESTORE then begin
          if UnlockDialog(LoggedUser, OnUnlock) then
            Application.UnhookMainWindow(UnlockHook)
          else
            UnlockHook := True;
        end;
    end;
  end;
end;

end.
