{*******************************************************}
{                                                       }
{         Delphi VCL Extensions (RX)                    }
{                                                       }
{         Copyright (c) 1995, 1996 AO ROSNO             }
{                                                       }
{*******************************************************}

unit Parsing;

interface

uses SysUtils;

type
  TRxMathParser = class
  private
    FCurPos: Cardinal;
    FParseText: string;
    function GetChar: Char;
    procedure NextChar;
    function GetNumber(var AValue: Double): Boolean;
    function GetConst(var AValue: Double): Boolean;
    function GetFunction(var AValue: ShortInt): Boolean;
    function Term: Double;
    function SubTerm: Double;
    function Calculate: Double;
  public
    ResValue: Double;
    function Exec(const AFormula: string): Boolean;
  end;

  ERxParserError = class(Exception);

implementation

uses RxTConst;

const
  SpecialChars = [#0..' ', '+', '-', '/', '*', ')'];
  nFunc = 12;
  FuncNames: array[1..nFunc] of string[6] = ('ARCTAN', 'COS', 'SIN', 'TAN',
    'ABS', 'EXP', 'LN', 'LOG', 'SQRT', 'SQR', 'INT', 'FRAC');

{ Parser errors }

procedure InvalidCondition(Str: Word);
begin
  raise ERxParserError.Create(LoadStr(Str));
end;

{ TRxMathParser }

function TRxMathParser.GetChar: Char;
begin
  Result := FParseText[FCurPos];
end;

procedure TRxMathParser.NextChar;
begin
  Inc(FCurPos);
end;

function TRxMathParser.GetNumber(var AValue: Double): Boolean;
var
  C: Char;
  SavePos: Cardinal;
  Value: Extended;
  Code: Integer;
  TmpStr: String;
begin
  Result := False;
  C := GetChar;
  SavePos := FCurPos;
  if C in ['0'..'9', '.', '+', '-'] then begin
    TmpStr := '';
    if (C = '.') then TmpStr := '0' + C else TmpStr := C;
    NextChar;
    C := GetChar;
    if (Length(TmpStr) = 1) and (TmpStr[1] in ['+', '-']) and
      (C = '.') then TmpStr := TmpStr + '0';
    while C in ['0'..'9', '.', 'E'] do begin
      TmpStr := TmpStr + C;
      if (C = 'E') then begin
        if (Length(TmpStr) > 1) and (TmpStr[Length(TmpStr) - 1] = '.') then
          Insert('0', TmpStr, Length(TmpStr));
        NextChar;
        C := GetChar;
        if (C in ['+', '-']) then begin
          TmpStr := TmpStr + C;
          NextChar;
        end;
      end
      else NextChar;
      C := GetChar;
    end;
    if (TmpStr[Length(TmpStr)] = '.') and (Pos('E', TmpStr) = 0) then
      TmpStr := TmpStr + '0';
{$IFDEF WIN32}
  {$HINTS OFF}
{$ENDIF}
    Val(TmpStr, Value, Code);
    if Code <> 0 then FCurPos := SavePos
    else Result := True;
  end;
  if Result and not (FParseText[FCurPos] in SpecialChars) then begin
    FCurPos := SavePos;
    Result := False;
  end;
  if Result then AValue := StrToFloat(TmpStr)
  else AValue := 0;
end;

function TRxMathParser.GetConst(var AValue: Double): Boolean;
begin
  Result := False;
  case FParseText[FCurPos] of
    'E':
      if FParseText[FCurPos + 1] in SpecialChars then 
      begin
        AValue := Exp(1);
        Inc(FCurPos);
        Result := True;
      end;
    'P':
      if (FParseText[FCurPos + 1] = 'I') and
        (FParseText[FCurPos + 1] in SpecialChars) then
      begin
        AValue := Pi;
        Inc(FCurPos, 2);
        Result := True;
      end;
  end
end;

function TRxMathParser.GetFunction(var AValue: Shortint): Boolean;
var
  I: Cardinal;
  TmpStr: String;
begin
  Result := False;
  AValue := 0;
  if FParseText[FCurPos] in ['A'..'Z'] then begin
    for I := 1 to nFunc do begin
      TmpStr := Copy(FParseText, FCurPos, Length(FuncNames[i]));
      if CompareStr(TmpStr, FuncNames[i]) = 0 then begin
        AValue := I;
        Break;
      end;
    end;
    if AValue > 0 then
      if FParseText[FCurPos + Length(TmpStr)] = '(' then begin
        Result := True;
        Inc(FCurPos, Length(TmpStr));
      end;
  end
end;

function TRxMathParser.Term: Double;
var
  Value: Double;
  NoFunc: ShortInt;
begin
  if FParseText[FCurPos] = '(' then begin
    Inc(FCurPos);
    Value := Calculate;
    if FParseText[FCurPos] <> ')' then InvalidCondition(SParseNotCramp);
    Inc(FCurPos);
  end
  else begin
    if not GetNumber(Value) then
      if not GetConst(Value) then
        if GetFunction(NoFunc) then begin
          Inc(FCurPos);
          Value := Calculate;
          try
            case NoFunc of
              1: Value := ArcTan(Value);
              2: Value := Cos(Value);
              3: Value := Sin(Value);
              4: if Cos(Value) = 0 then InvalidCondition(SParseDivideByZero)
                 else Value := Sin(Value) / Cos(Value);
              5: Value := Abs(Value);
              6: Value := Exp(Value);
              7: if Value <= 0 then InvalidCondition(SParseLogError)
                 else Value := Ln(Value);
              8: if Value <= 0 then InvalidCondition(SParseLogError)
                 else Value := Ln(Value) / Ln(10);
              9: if Value < 0 then InvalidCondition(SParseSqrError)
                 else Value := Sqrt(Value);
              10: Value := Sqr(Value);
              11: Value := Round(Value);
              12: Value := Frac(Value);
            end;
          except
            on E: ERxParserError do raise
            else InvalidCondition(SParseInvalidFloatOperation);
          end;
          if FParseText[FCurPos] <> ')' then InvalidCondition(SParseNotCramp);
          Inc(FCurPos);
        end
        else InvalidCondition(SParseSyntaxError);
  end;
  Term := Value;
end;

function TRxMathParser.SubTerm: Double;
var
  Value: Double;
begin
  Value := Term;
  while FParseText[FCurPos] in ['*','/'] do begin
    Inc(FCurPos);
    if FParseText[FCurPos-1] = '*' then
      Value := Value * Term
    else
    try
      Value := Value / Term;
    except
      InvalidCondition(SParseDivideByZero);
    end;
  end;
  SubTerm := Value;
end;

function TRxMathParser.Calculate: Double;
var
  Value: Double;
begin
  Value := SubTerm;
  while FParseText[FCurPos] in ['+','-'] do begin
    Inc(FCurPos);
    if FParseText[FCurPos - 1] = '+' then Value := Value + SubTerm
    else Value := Value - SubTerm;
  end;
  if not (FParseText[FCurPos] in [#0, ')', '>', '<', '=', ',']) then
    InvalidCondition(SParseSyntaxError);
  Result := Value;
end;

function TRxMathParser.Exec(const AFormula: string): Boolean;
var
  I, J: Integer;
begin
  Result := False;
  J := 0;
  FParseText := '';
  for I := 1 to Length(AFormula) do begin
    case AFormula[I] of
      '(': Inc(J);
      ')': Dec(J);
    end;
    if AFormula[I] > ' ' then FParseText := FParseText + UpCase(AFormula[I]);
  end;
  if J = 0 then begin
    FCurPos := 1;
    FParseText := FParseText + #0;
    ResValue := Calculate;
    Result := True;
  end
  else InvalidCondition(SParseNotCramp);
end;

end.
